# Copyright (C) 2018 by Kevin L. Mitchell <klmitch@mit.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest
import webob.exc

from micropath import controller
from micropath import elements


class ExceptionForTest(Exception):
    pass


class TestControllerMeta(object):
    def test_base(self, mocker):
        mock_Root = mocker.patch.object(controller.elements, 'Root')

        result = controller.ControllerMeta(
            'TestController', (controller.Controller,), {},
        )

        assert result._micropath_root == mock_Root.return_value
        assert result._micropath_delegations == []
        mock_Root.assert_called_once_with()

    def test_alt(self, mocker):
        mock_Root = mocker.patch.object(controller.elements, 'Root')
        namespace = {
            'elem1': mocker.Mock(spec=elements.Path),
            'elem2': mocker.Mock(spec=elements.Binding),
            'deleg1': mocker.Mock(spec=elements.Delegation, element=None),
            'deleg2': mocker.Mock(spec=elements.Delegation, element='elem'),
            'func': mocker.Mock(
                _micropath_methods=[
                    mocker.Mock(spec=elements.Method),
                    mocker.Mock(spec=elements.Method),
                ],
                _micropath_handler=False,
                _micropath_elem=None,
            ),
            'handler1': mocker.Mock(
                _micropath_methods=[],
                _micropath_handler=True,
                _micropath_elem=None,
            ),
            'handler2': mocker.Mock(
                _micropath_methods=[],
                _micropath_handler=True,
                _micropath_elem='elem',
            ),
        }

        result = controller.ControllerMeta(
            'TestController', (controller.Controller,), namespace,
        )

        assert result._micropath_root == mock_Root.return_value
        assert result._micropath_delegations == [
            namespace['deleg1'],
            namespace['deleg2'],
        ]
        assert result.func._micropath_elem is None
        assert result.handler1._micropath_elem == mock_Root.return_value
        assert result.handler2._micropath_elem == 'elem'
        mock_Root.assert_called_once_with()
        mock_Root.return_value.mount.assert_called_once_with(
            namespace['deleg1'],
        )
        mock_Root.return_value.add_elem.assert_has_calls([
            mocker.call(namespace['deleg2'].element, 'deleg2'),
            mocker.call(namespace['elem1'], 'elem1'),
            mocker.call(namespace['elem2'], 'elem2'),
            mocker.call(namespace['func']._micropath_methods[0], 'func'),
            mocker.call(namespace['func']._micropath_methods[1], 'func'),
        ], any_order=True)
        assert mock_Root.return_value.add_elem.call_count == 5


class TestController(object):
    def test_init(self, mocker):
        delegations = [
            mocker.Mock(),
            mocker.Mock(),
            mocker.Mock(),
        ]
        mocker.patch.object(
            controller.Controller, '_micropath_delegations', delegations,
        )

        result = controller.Controller()

        assert isinstance(result, controller.Controller)
        for deleg in delegations:
            deleg.get.assert_called_once_with(result)

    def check_injector(self, obj, req, mocker, mock_exc_info, **kwargs):
        injector = req.injector.cleanup.return_value.__enter__.return_value
        injector.update.assert_called_once_with(kwargs)
        injector.__setitem__.assert_has_calls([
            mocker.call('request', req),
            mocker.call('root_controller', obj),
        ])
        assert injector.__setitem__.call_count == 2

        mock_micropath_request_error = mocker.patch.object(
            obj, 'micropath_request_error',
            return_value=ExceptionForTest(),
        )
        keys = set()
        for pos, _kw in injector.set_deferred.call_args_list:
            # Get the call arguments
            key = pos[0]
            func = pos[1]
            real_key = (
                controller.Controller.micropath_request_attrs[key] or key
            )

            # Check what happens when the function is called
            assert func(req) == getattr(req, real_key)

            # Check what happens if the attribute doesn't exist
            delattr(req, real_key)
            with pytest.raises(ExceptionForTest):
                func(req)
            mock_micropath_request_error.assert_called_once_with(
                req, real_key, mock_exc_info.return_value,
            )
            mock_micropath_request_error.reset_mock()

            # We've tested for this key
            keys.add(pos[0])

        assert keys == set(controller.Controller.micropath_request_attrs)

    def test_call_base(self, mocker):
        req = mocker.MagicMock(charset='utf-8', urlvars={'a': 1})
        inj = req.injector.cleanup.return_value.__enter__.return_value
        mock_Request = mocker.patch.object(
            controller.Controller, 'micropath_request',
            return_value=req,
        )
        mock_micropath_prepare_injector = mocker.patch.object(
            controller.Controller, 'micropath_prepare_injector',
        )
        mock_micropath_dispatch = mocker.patch.object(
            controller.Controller, '_micropath_dispatch',
        )
        base_resp = mock_micropath_dispatch.return_value
        mock_exc_info = mocker.patch.object(
            controller.sys, 'exc_info',
        )
        mock_micropath_server_error = mocker.patch.object(
            controller.Controller, 'micropath_server_error',
        )
        mocker.patch.object(
            controller.traceback, 'format_exc',
            return_value='exception',
        )
        mock_HTTPInternalServerError = mocker.patch.object(
            controller.webob.exc, 'HTTPInternalServerError',
        )
        obj = controller.Controller()

        result = obj('environ', 'start_response')

        assert req.response == req.ResponseClass.return_value
        resp = req.response.merge_cookies.return_value
        assert result == resp.return_value
        mock_Request.assert_called_once_with('environ')
        req.ResponseClass.assert_called_once_with()
        req.injector.cleanup.assert_called_once_with()
        mock_micropath_prepare_injector.assert_called_once_with(req, inj)
        mock_micropath_dispatch.assert_called_once_with(req, inj)
        mock_exc_info.assert_not_called()
        mock_micropath_server_error.assert_not_called()
        mock_HTTPInternalServerError.assert_not_called()
        req.response.write.assert_not_called()
        req.response.merge_cookies.assert_called_once_with(base_resp)
        resp.assert_called_once_with('environ', 'start_response')
        self.check_injector(obj, req, mocker, mock_exc_info, a=1)

    def test_call_return_none(self, mocker):
        req = mocker.MagicMock(charset='utf-8', urlvars={'a': 1})
        inj = req.injector.cleanup.return_value.__enter__.return_value
        mock_Request = mocker.patch.object(
            controller.Controller, 'micropath_request',
            return_value=req,
        )
        mock_micropath_prepare_injector = mocker.patch.object(
            controller.Controller, 'micropath_prepare_injector',
        )
        mock_micropath_dispatch = mocker.patch.object(
            controller.Controller, '_micropath_dispatch',
            return_value=None,
        )
        mock_exc_info = mocker.patch.object(
            controller.sys, 'exc_info',
        )
        mock_micropath_server_error = mocker.patch.object(
            controller.Controller, 'micropath_server_error',
        )
        mocker.patch.object(
            controller.traceback, 'format_exc',
            return_value='exception',
        )
        mock_HTTPInternalServerError = mocker.patch.object(
            controller.webob.exc, 'HTTPInternalServerError',
        )
        obj = controller.Controller()

        result = obj('environ', 'start_response')

        assert req.response == req.ResponseClass.return_value
        resp = req.response
        assert result == resp.return_value
        mock_Request.assert_called_once_with('environ')
        req.ResponseClass.assert_called_once_with()
        req.injector.cleanup.assert_called_once_with()
        mock_micropath_prepare_injector.assert_called_once_with(req, inj)
        mock_micropath_dispatch.assert_called_once_with(req, inj)
        mock_exc_info.assert_not_called()
        mock_micropath_server_error.assert_not_called()
        mock_HTTPInternalServerError.assert_not_called()
        req.response.write.assert_not_called()
        req.response.merge_cookies.assert_not_called()
        resp.assert_called_once_with('environ', 'start_response')
        self.check_injector(obj, req, mocker, mock_exc_info, a=1)

    def test_call_return_text(self, mocker):
        req = mocker.MagicMock(charset='utf-8', urlvars={'a': 1})
        inj = req.injector.cleanup.return_value.__enter__.return_value
        mock_Request = mocker.patch.object(
            controller.Controller, 'micropath_request',
            return_value=req,
        )
        mock_micropath_prepare_injector = mocker.patch.object(
            controller.Controller, 'micropath_prepare_injector',
        )
        mock_micropath_dispatch = mocker.patch.object(
            controller.Controller, '_micropath_dispatch',
            return_value=u'this\u2026is a test',
        )
        mock_exc_info = mocker.patch.object(
            controller.sys, 'exc_info',
        )
        mock_micropath_server_error = mocker.patch.object(
            controller.Controller, 'micropath_server_error',
        )
        mocker.patch.object(
            controller.traceback, 'format_exc',
            return_value='exception',
        )
        mock_HTTPInternalServerError = mocker.patch.object(
            controller.webob.exc, 'HTTPInternalServerError',
        )
        obj = controller.Controller()

        result = obj('environ', 'start_response')

        assert req.response == req.ResponseClass.return_value
        resp = req.response
        assert result == resp.return_value
        mock_Request.assert_called_once_with('environ')
        req.ResponseClass.assert_called_once_with()
        req.injector.cleanup.assert_called_once_with()
        mock_micropath_prepare_injector.assert_called_once_with(req, inj)
        mock_micropath_dispatch.assert_called_once_with(req, inj)
        mock_exc_info.assert_not_called()
        mock_micropath_server_error.assert_not_called()
        mock_HTTPInternalServerError.assert_not_called()
        req.response.write.assert_called_once_with(
            b'this\xe2\x80\xa6is a test',
        )
        req.response.merge_cookies.assert_not_called()
        resp.assert_called_once_with('environ', 'start_response')
        self.check_injector(obj, req, mocker, mock_exc_info, a=1)

    def test_call_return_bytes(self, mocker):
        req = mocker.MagicMock(charset='utf-8', urlvars={'a': 1})
        inj = req.injector.cleanup.return_value.__enter__.return_value
        mock_Request = mocker.patch.object(
            controller.Controller, 'micropath_request',
            return_value=req,
        )
        mock_micropath_prepare_injector = mocker.patch.object(
            controller.Controller, 'micropath_prepare_injector',
        )
        mock_micropath_dispatch = mocker.patch.object(
            controller.Controller, '_micropath_dispatch',
            return_value=b'this\xe2\x80\xa6is a test',
        )
        mock_exc_info = mocker.patch.object(
            controller.sys, 'exc_info',
        )
        mock_micropath_server_error = mocker.patch.object(
            controller.Controller, 'micropath_server_error',
        )
        mocker.patch.object(
            controller.traceback, 'format_exc',
            return_value='exception',
        )
        mock_HTTPInternalServerError = mocker.patch.object(
            controller.webob.exc, 'HTTPInternalServerError',
        )
        obj = controller.Controller()

        result = obj('environ', 'start_response')

        assert req.response == req.ResponseClass.return_value
        resp = req.response
        assert result == resp.return_value
        mock_Request.assert_called_once_with('environ')
        req.ResponseClass.assert_called_once_with()
        req.injector.cleanup.assert_called_once_with()
        mock_micropath_prepare_injector.assert_called_once_with(req, inj)
        mock_micropath_dispatch.assert_called_once_with(req, inj)
        mock_exc_info.assert_not_called()
        mock_micropath_server_error.assert_not_called()
        mock_HTTPInternalServerError.assert_not_called()
        req.response.write.assert_called_once_with(
            b'this\xe2\x80\xa6is a test',
        )
        req.response.merge_cookies.assert_not_called()
        resp.assert_called_once_with('environ', 'start_response')
        self.check_injector(obj, req, mocker, mock_exc_info, a=1)

    def test_call_http_exception(self, mocker):
        req = mocker.MagicMock(charset='utf-8', urlvars={'a': 1})
        inj = req.injector.cleanup.return_value.__enter__.return_value
        mock_Request = mocker.patch.object(
            controller.Controller, 'micropath_request',
            return_value=req,
        )
        mock_micropath_prepare_injector = mocker.patch.object(
            controller.Controller, 'micropath_prepare_injector',
        )
        exc = webob.exc.HTTPException('test', 'wsgi_response')
        mock_micropath_dispatch = mocker.patch.object(
            controller.Controller, '_micropath_dispatch',
            side_effect=exc,
        )
        mock_exc_info = mocker.patch.object(
            controller.sys, 'exc_info',
        )
        mock_micropath_server_error = mocker.patch.object(
            controller.Controller, 'micropath_server_error',
        )
        mocker.patch.object(
            controller.traceback, 'format_exc',
            return_value='exception',
        )
        mock_HTTPInternalServerError = mocker.patch.object(
            controller.webob.exc, 'HTTPInternalServerError',
        )
        obj = controller.Controller()

        result = obj('environ', 'start_response')

        assert req.response == req.ResponseClass.return_value
        resp = req.response.merge_cookies.return_value
        assert result == resp.return_value
        mock_Request.assert_called_once_with('environ')
        req.ResponseClass.assert_called_once_with()
        req.injector.cleanup.assert_called_once_with()
        mock_micropath_prepare_injector.assert_called_once_with(req, inj)
        mock_micropath_dispatch.assert_called_once_with(req, inj)
        mock_exc_info.assert_not_called()
        mock_micropath_server_error.assert_not_called()
        mock_HTTPInternalServerError.assert_not_called()
        req.response.write.assert_not_called()
        req.response.merge_cookies.assert_called_once_with(exc)
        resp.assert_called_once_with('environ', 'start_response')
        self.check_injector(obj, req, mocker, mock_exc_info, a=1)

    def test_call_exceptionfortest(self, mocker):
        req = mocker.MagicMock(charset='utf-8', urlvars={'a': 1})
        inj = req.injector.cleanup.return_value.__enter__.return_value
        mock_Request = mocker.patch.object(
            controller.Controller, 'micropath_request',
            return_value=req,
        )
        mock_micropath_prepare_injector = mocker.patch.object(
            controller.Controller, 'micropath_prepare_injector',
        )
        mock_micropath_dispatch = mocker.patch.object(
            controller.Controller, '_micropath_dispatch',
            side_effect=ExceptionForTest('test'),
        )
        mock_exc_info = mocker.patch.object(
            controller.sys, 'exc_info',
        )
        mock_micropath_server_error = mocker.patch.object(
            controller.Controller, 'micropath_server_error',
        )
        mocker.patch.object(
            controller.traceback, 'format_exc',
            return_value='exception',
        )
        mock_HTTPInternalServerError = mocker.patch.object(
            controller.webob.exc, 'HTTPInternalServerError',
        )
        obj = controller.Controller()

        result = obj('environ', 'start_response')

        assert req.response == req.ResponseClass.return_value
        resp = req.response.merge_cookies.return_value
        assert result == resp.return_value
        mock_Request.assert_called_once_with('environ')
        req.ResponseClass.assert_called_once_with()
        req.injector.cleanup.assert_called_once_with()
        mock_micropath_prepare_injector.assert_called_once_with(req, inj)
        mock_micropath_dispatch.assert_called_once_with(req, inj)
        mock_exc_info.assert_called_once_with()
        mock_micropath_server_error.assert_called_once_with(
            req, mock_exc_info.return_value,
        )
        base_resp = mock_micropath_server_error.return_value
        mock_HTTPInternalServerError.assert_not_called()
        req.response.write.assert_not_called()
        req.response.merge_cookies.assert_called_once_with(base_resp)
        resp.assert_called_once_with('environ', 'start_response')
        self.check_injector(obj, req, mocker, mock_exc_info, a=1)

    def test_call_last_resort_exception(self, mocker):
        req = mocker.MagicMock(charset='utf-8', urlvars={'a': 1})
        inj = req.injector.cleanup.return_value.__enter__.return_value
        mock_Request = mocker.patch.object(
            controller.Controller, 'micropath_request',
            return_value=req,
        )
        mock_micropath_prepare_injector = mocker.patch.object(
            controller.Controller, 'micropath_prepare_injector',
        )
        mock_micropath_dispatch = mocker.patch.object(
            controller.Controller, '_micropath_dispatch',
            side_effect=ExceptionForTest('test'),
        )
        mock_exc_info = mocker.patch.object(
            controller.sys, 'exc_info',
        )
        mock_micropath_server_error = mocker.patch.object(
            controller.Controller, 'micropath_server_error',
            side_effect=ExceptionForTest('test'),
        )
        mocker.patch.object(
            controller.traceback, 'format_exc',
            return_value='exception',
        )
        mock_HTTPInternalServerError = mocker.patch.object(
            controller.webob.exc, 'HTTPInternalServerError',
        )
        obj = controller.Controller()

        result = obj('environ', 'start_response')

        assert req.response == req.ResponseClass.return_value
        resp = req.response.merge_cookies.return_value
        assert result == resp.return_value
        mock_Request.assert_called_once_with('environ')
        req.ResponseClass.assert_called_once_with()
        req.injector.cleanup.assert_called_once_with()
        mock_micropath_prepare_injector.assert_called_once_with(req, inj)
        mock_micropath_dispatch.assert_called_once_with(req, inj)
        mock_exc_info.assert_called_once_with()
        mock_micropath_server_error.assert_called_once_with(
            req, mock_exc_info.return_value,
        )
        mock_HTTPInternalServerError.assert_called_once_with(None)
        base_resp = mock_HTTPInternalServerError.return_value
        req.response.write.assert_not_called()
        req.response.merge_cookies.assert_called_once_with(base_resp)
        resp.assert_called_once_with('environ', 'start_response')
        self.check_injector(obj, req, mocker, mock_exc_info, a=1)

    def test_call_last_resort_exception_debug(self, mocker):
        req = mocker.MagicMock(charset='utf-8', urlvars={'a': 1})
        inj = req.injector.cleanup.return_value.__enter__.return_value
        mock_Request = mocker.patch.object(
            controller.Controller, 'micropath_request',
            return_value=req,
        )
        mock_micropath_prepare_injector = mocker.patch.object(
            controller.Controller, 'micropath_prepare_injector',
        )
        exc = ExceptionForTest('test')
        mock_micropath_dispatch = mocker.patch.object(
            controller.Controller, '_micropath_dispatch',
            side_effect=exc,
        )
        mock_exc_info = mocker.patch.object(
            controller.sys, 'exc_info',
        )
        mock_micropath_server_error = mocker.patch.object(
            controller.Controller, 'micropath_server_error',
            side_effect=exc,
        )
        mocker.patch.object(
            controller.traceback, 'format_exc',
            return_value='exception',
        )
        mock_HTTPInternalServerError = mocker.patch.object(
            controller.webob.exc, 'HTTPInternalServerError',
        )
        obj = controller.Controller()
        obj.micropath_debug = True

        result = obj('environ', 'start_response')

        assert req.response == req.ResponseClass.return_value
        resp = req.response.merge_cookies.return_value
        assert result == resp.return_value
        mock_Request.assert_called_once_with('environ')
        req.ResponseClass.assert_called_once_with()
        req.injector.cleanup.assert_called_once_with()
        mock_micropath_prepare_injector.assert_called_once_with(req, inj)
        mock_micropath_dispatch.assert_called_once_with(req, inj)
        mock_exc_info.assert_called_once_with()
        mock_micropath_server_error.assert_called_once_with(
            req, mock_exc_info.return_value,
        )
        mock_HTTPInternalServerError.assert_called_once_with('exception')
        base_resp = mock_HTTPInternalServerError.return_value
        req.response.write.assert_not_called()
        req.response.merge_cookies.assert_called_once_with(base_resp)
        resp.assert_called_once_with('environ', 'start_response')
        self.check_injector(obj, req, mocker, mock_exc_info, a=1)

    def test_micropath_dispatch_call_func(self, mocker):
        mock_wants = mocker.patch.object(
            controller.injector, 'wants',
            return_value=False,
        )
        mock_micropath_not_found = mocker.patch.object(
            controller.Controller, 'micropath_not_found',
        )
        mock_micropath_not_implemented = mocker.patch.object(
            controller.Controller, 'micropath_not_implemented',
        )
        mock_micropath_options = mocker.patch.object(
            controller.Controller, 'micropath_options',
        )
        elem = mocker.Mock(methods={})
        mock_micropath_resolve = mocker.patch.object(
            controller.Controller, '_micropath_resolve',
            return_value=(elem, False),
        )
        mock_micropath_delegation = mocker.patch.object(
            controller.Controller, '_micropath_delegation',
            return_value=('func', None),
        )
        mock_micropath_methods = mocker.patch.object(
            controller.Controller, '_micropath_methods',
            return_value=set(['HEAD', 'GET', 'POST']),
        )
        req = mocker.Mock(
            path_info='path_info',
            method='GET',
        )
        inj = mocker.Mock()
        obj = controller.Controller()

        result = obj._micropath_dispatch(req, inj)

        assert result == inj.return_value
        mock_wants.assert_not_called()
        mock_micropath_not_found.assert_not_called()
        mock_micropath_not_implemented.assert_not_called()
        mock_micropath_options.assert_not_called()
        mock_micropath_resolve.assert_called_once_with(req, inj)
        mock_micropath_delegation.assert_called_once_with(req, elem)
        mock_micropath_methods.assert_not_called()
        inj.assert_called_once_with('func', obj)

    def test_micropath_dispatch_call_func_wants_path_info(self, mocker):
        mock_wants = mocker.patch.object(
            controller.injector, 'wants',
            return_value=True,
        )
        mock_micropath_not_found = mocker.patch.object(
            controller.Controller, 'micropath_not_found',
        )
        mock_micropath_not_implemented = mocker.patch.object(
            controller.Controller, 'micropath_not_implemented',
        )
        mock_micropath_options = mocker.patch.object(
            controller.Controller, 'micropath_options',
        )
        elem = mocker.Mock(methods={})
        mock_micropath_resolve = mocker.patch.object(
            controller.Controller, '_micropath_resolve',
            return_value=(elem, True),
        )
        mock_micropath_delegation = mocker.patch.object(
            controller.Controller, '_micropath_delegation',
            return_value=('func', None),
        )
        mock_micropath_methods = mocker.patch.object(
            controller.Controller, '_micropath_methods',
            return_value=set(['HEAD', 'GET', 'POST']),
        )
        req = mocker.Mock(
            path_info='path_info',
            method='GET',
        )
        inj = mocker.Mock()
        obj = controller.Controller()

        result = obj._micropath_dispatch(req, inj)

        assert result == inj.return_value
        mock_wants.assert_called_once_with('func', 'path_info')
        mock_micropath_not_found.assert_not_called()
        mock_micropath_not_implemented.assert_not_called()
        mock_micropath_options.assert_not_called()
        mock_micropath_resolve.assert_called_once_with(req, inj)
        mock_micropath_delegation.assert_called_once_with(req, elem)
        mock_micropath_methods.assert_not_called()
        inj.assert_called_once_with('func', obj)

    def test_micropath_dispatch_call_func_doesnt_want_path_info(self, mocker):
        mock_wants = mocker.patch.object(
            controller.injector, 'wants',
            return_value=False,
        )
        mock_micropath_not_found = mocker.patch.object(
            controller.Controller, 'micropath_not_found',
        )
        mock_micropath_not_implemented = mocker.patch.object(
            controller.Controller, 'micropath_not_implemented',
        )
        mock_micropath_options = mocker.patch.object(
            controller.Controller, 'micropath_options',
        )
        elem = mocker.Mock(methods={})
        mock_micropath_resolve = mocker.patch.object(
            controller.Controller, '_micropath_resolve',
            return_value=(elem, True),
        )
        mock_micropath_delegation = mocker.patch.object(
            controller.Controller, '_micropath_delegation',
            return_value=('func', None),
        )
        mock_micropath_methods = mocker.patch.object(
            controller.Controller, '_micropath_methods',
            return_value=set(['HEAD', 'GET', 'POST']),
        )
        req = mocker.Mock(
            path_info='path_info',
            method='GET',
        )
        inj = mocker.Mock()
        obj = controller.Controller()

        result = obj._micropath_dispatch(req, inj)

        assert result == mock_micropath_not_found.return_value
        mock_wants.assert_called_once_with('func', 'path_info')
        mock_micropath_not_found.assert_called_once_with(req, 'path_info')
        mock_micropath_not_implemented.assert_not_called()
        mock_micropath_options.assert_not_called()
        mock_micropath_resolve.assert_called_once_with(req, inj)
        mock_micropath_delegation.assert_called_once_with(req, elem)
        mock_micropath_methods.assert_not_called()
        inj.assert_not_called()

    def test_micropath_dispatch_call_delegation(self, mocker):
        mock_wants = mocker.patch.object(
            controller.injector, 'wants',
            return_value=False,
        )
        mock_micropath_not_found = mocker.patch.object(
            controller.Controller, 'micropath_not_found',
        )
        mock_micropath_not_implemented = mocker.patch.object(
            controller.Controller, 'micropath_not_implemented',
        )
        mock_micropath_options = mocker.patch.object(
            controller.Controller, 'micropath_options',
        )
        elem = mocker.Mock(methods={})
        mock_micropath_resolve = mocker.patch.object(
            controller.Controller, '_micropath_resolve',
            return_value=(elem, False),
        )
        delegation = mocker.Mock()
        delegation_obj = delegation.get.return_value
        mock_micropath_delegation = mocker.patch.object(
            controller.Controller, '_micropath_delegation',
            return_value=(None, delegation),
        )
        mock_micropath_methods = mocker.patch.object(
            controller.Controller, '_micropath_methods',
            return_value=set(['HEAD', 'GET', 'POST']),
        )
        req = mocker.Mock(
            path_info='path_info',
            method='GET',
        )
        inj = mocker.Mock()
        obj = controller.Controller()

        result = obj._micropath_dispatch(req, inj)

        assert result == delegation_obj._micropath_dispatch.return_value
        mock_wants.assert_not_called()
        mock_micropath_not_found.assert_not_called()
        mock_micropath_not_implemented.assert_not_called()
        mock_micropath_options.assert_not_called()
        mock_micropath_resolve.assert_called_once_with(req, inj)
        mock_micropath_delegation.assert_called_once_with(req, elem)
        mock_micropath_methods.assert_not_called()
        delegation.get.assert_called_once_with(obj)
        delegation_obj._micropath_dispatch.assert_called_once_with(req, inj)
        inj.assert_not_called()

    def test_micropath_dispatch_default_options(self, mocker):
        mock_wants = mocker.patch.object(
            controller.injector, 'wants',
            return_value=False,
        )
        mock_micropath_not_found = mocker.patch.object(
            controller.Controller, 'micropath_not_found',
        )
        mock_micropath_not_implemented = mocker.patch.object(
            controller.Controller, 'micropath_not_implemented',
        )
        mock_micropath_options = mocker.patch.object(
            controller.Controller, 'micropath_options',
        )
        elem = mocker.Mock(methods={'GET': 'getter'})
        mock_micropath_resolve = mocker.patch.object(
            controller.Controller, '_micropath_resolve',
            return_value=(elem, False),
        )
        mock_micropath_delegation = mocker.patch.object(
            controller.Controller, '_micropath_delegation',
            return_value=(None, None),
        )
        mock_micropath_methods = mocker.patch.object(
            controller.Controller, '_micropath_methods',
            return_value=set(['HEAD', 'GET', 'POST']),
        )
        req = mocker.Mock(
            path_info='path_info',
            method='OPTIONS',
        )
        inj = mocker.Mock()
        obj = controller.Controller()

        result = obj._micropath_dispatch(req, inj)

        assert result == mock_micropath_options.return_value
        mock_wants.assert_not_called()
        mock_micropath_not_found.assert_not_called()
        mock_micropath_not_implemented.assert_not_called()
        mock_micropath_options.assert_called_once_with(
            req, ['GET', 'HEAD', 'POST'],
        )
        mock_micropath_resolve.assert_called_once_with(req, inj)
        mock_micropath_delegation.assert_called_once_with(req, elem)
        mock_micropath_methods.assert_called_once_with(elem)
        inj.assert_not_called()

    def test_micropath_dispatch_default_options_no_methods(self, mocker):
        mock_wants = mocker.patch.object(
            controller.injector, 'wants',
            return_value=False,
        )
        mock_micropath_not_found = mocker.patch.object(
            controller.Controller, 'micropath_not_found',
        )
        mock_micropath_not_implemented = mocker.patch.object(
            controller.Controller, 'micropath_not_implemented',
        )
        mock_micropath_options = mocker.patch.object(
            controller.Controller, 'micropath_options',
        )
        elem = mocker.Mock(methods={})
        mock_micropath_resolve = mocker.patch.object(
            controller.Controller, '_micropath_resolve',
            return_value=(elem, False),
        )
        mock_micropath_delegation = mocker.patch.object(
            controller.Controller, '_micropath_delegation',
            return_value=(None, None),
        )
        mock_micropath_methods = mocker.patch.object(
            controller.Controller, '_micropath_methods',
            return_value=set(['HEAD', 'GET', 'POST']),
        )
        req = mocker.Mock(
            path_info='path_info',
            method='OPTIONS',
        )
        inj = mocker.Mock()
        obj = controller.Controller()

        result = obj._micropath_dispatch(req, inj)

        assert result == mock_micropath_not_found.return_value
        mock_wants.assert_not_called()
        mock_micropath_not_found.assert_called_once_with(req, 'path_info')
        mock_micropath_not_implemented.assert_not_called()
        mock_micropath_options.assert_not_called()
        mock_micropath_resolve.assert_called_once_with(req, inj)
        mock_micropath_delegation.assert_called_once_with(req, elem)
        mock_micropath_methods.assert_not_called()
        inj.assert_not_called()

    def test_micropath_dispatch_not_implemented(self, mocker):
        mock_wants = mocker.patch.object(
            controller.injector, 'wants',
            return_value=False,
        )
        mock_micropath_not_found = mocker.patch.object(
            controller.Controller, 'micropath_not_found',
        )
        mock_micropath_not_implemented = mocker.patch.object(
            controller.Controller, 'micropath_not_implemented',
        )
        mock_micropath_options = mocker.patch.object(
            controller.Controller, 'micropath_options',
        )
        elem = mocker.Mock(methods={'GET': 'getter'})
        mock_micropath_resolve = mocker.patch.object(
            controller.Controller, '_micropath_resolve',
            return_value=(elem, False),
        )
        mock_micropath_delegation = mocker.patch.object(
            controller.Controller, '_micropath_delegation',
            return_value=(None, None),
        )
        mock_micropath_methods = mocker.patch.object(
            controller.Controller, '_micropath_methods',
            return_value=set(['HEAD', 'GET', 'POST']),
        )
        req = mocker.Mock(
            path_info='path_info',
            method='GET',
        )
        inj = mocker.Mock()
        obj = controller.Controller()

        result = obj._micropath_dispatch(req, inj)

        assert result == mock_micropath_not_implemented.return_value
        mock_wants.assert_not_called()
        mock_micropath_not_found.assert_not_called()
        mock_micropath_not_implemented.assert_called_once_with(req, 'GET')
        mock_micropath_options.assert_not_called()
        mock_micropath_resolve.assert_called_once_with(req, inj)
        mock_micropath_delegation.assert_called_once_with(req, elem)
        mock_micropath_methods.assert_not_called()
        inj.assert_not_called()

    def test_micropath_dispatch_no_methods(self, mocker):
        mock_wants = mocker.patch.object(
            controller.injector, 'wants',
            return_value=False,
        )
        mock_micropath_not_found = mocker.patch.object(
            controller.Controller, 'micropath_not_found',
        )
        mock_micropath_not_implemented = mocker.patch.object(
            controller.Controller, 'micropath_not_implemented',
        )
        mock_micropath_options = mocker.patch.object(
            controller.Controller, 'micropath_options',
        )
        elem = mocker.Mock(methods={})
        mock_micropath_resolve = mocker.patch.object(
            controller.Controller, '_micropath_resolve',
            return_value=(elem, False),
        )
        mock_micropath_delegation = mocker.patch.object(
            controller.Controller, '_micropath_delegation',
            return_value=(None, None),
        )
        mock_micropath_methods = mocker.patch.object(
            controller.Controller, '_micropath_methods',
            return_value=set(['HEAD', 'GET', 'POST']),
        )
        req = mocker.Mock(
            path_info='path_info',
            method='GET',
        )
        inj = mocker.Mock()
        obj = controller.Controller()

        result = obj._micropath_dispatch(req, inj)

        assert result == mock_micropath_not_found.return_value
        mock_wants.assert_not_called()
        mock_micropath_not_found.assert_called_once_with(req, 'path_info')
        mock_micropath_not_implemented.assert_not_called()
        mock_micropath_options.assert_not_called()
        mock_micropath_resolve.assert_called_once_with(req, inj)
        mock_micropath_delegation.assert_called_once_with(req, elem)
        mock_micropath_methods.assert_not_called()
        inj.assert_not_called()

    def test_micropath_resolve_base(self, mocker):
        elems = {
            None: mocker.Mock(t_paths=['a'], t_bindings=None, skip=False),
            'a': mocker.Mock(t_paths=[], t_bindings='b', skip=False),
            'b': mocker.Mock(t_paths=['c'], t_bindings=None, skip=False),
            'c': mocker.Mock(t_paths=[], t_bindings='d', skip=False),
            'd': mocker.Mock(t_paths=[], t_bindings=None, skip=False),
        }
        for name, elem in elems.items():
            elem.ident = name
            if elem.skip:
                elem.validate.side_effect = elements.SkipBinding()
            elem.paths = {x: elems[x] for x in elem.t_paths}
            elem.bindings = (
                None if elem.t_bindings is None else elems[elem.t_bindings]
            )
        mocker.patch.object(
            controller.Controller, '_micropath_root', elems[None],
        )
        req = mocker.Mock(**{
            'path_info_peek.side_effect': ['a', '1', 'c', '2', ''],
            'urlvars': {},
        })
        inj = {}
        obj = controller.Controller()

        result = obj._micropath_resolve(req, inj)

        assert result == (elems['d'], False)
        url_vars = {
            'b': elems['b'].validate.return_value,
            'd': elems['d'].validate.return_value,
        }
        assert inj == url_vars
        assert req.urlvars == url_vars
        elems['b'].validate.assert_called_once_with(obj, inj, '1')
        elems['d'].validate.assert_called_once_with(obj, inj, '2')

    def test_micropath_resolve_injector_conflict(self, mocker):
        elems = {
            None: mocker.Mock(t_paths=['a'], t_bindings=None, skip=False),
            'a': mocker.Mock(t_paths=[], t_bindings='b', skip=False),
            'b': mocker.Mock(t_paths=['c'], t_bindings=None, skip=False),
            'c': mocker.Mock(t_paths=[], t_bindings='d', skip=False),
            'd': mocker.Mock(t_paths=[], t_bindings=None, skip=False),
        }
        for name, elem in elems.items():
            elem.ident = name
            if elem.skip:
                elem.validate.side_effect = elements.SkipBinding()
            elem.paths = {x: elems[x] for x in elem.t_paths}
            elem.bindings = (
                None if elem.t_bindings is None else elems[elem.t_bindings]
            )
        mocker.patch.object(
            controller.Controller, '_micropath_root', elems[None],
        )
        req = mocker.Mock(**{
            'path_info_peek.side_effect': ['a', '1', 'c', '2', ''],
            'urlvars': {},
        })
        inj = {'d': 'x'}
        obj = controller.Controller()

        result = obj._micropath_resolve(req, inj)

        assert result == (elems['d'], False)
        assert inj == {
            'b': elems['b'].validate.return_value,
            'd': 'x',
        }
        assert req.urlvars == {
            'b': elems['b'].validate.return_value,
            'd': elems['d'].validate.return_value,
        }
        elems['b'].validate.assert_called_once_with(obj, inj, '1')
        elems['d'].validate.assert_called_once_with(obj, inj, '2')

    def test_micropath_resolve_skip_last(self, mocker):
        elems = {
            None: mocker.Mock(t_paths=['a'], t_bindings=None, skip=False),
            'a': mocker.Mock(t_paths=[], t_bindings='b', skip=False),
            'b': mocker.Mock(t_paths=['c'], t_bindings=None, skip=False),
            'c': mocker.Mock(t_paths=[], t_bindings='d', skip=False),
            'd': mocker.Mock(t_paths=[], t_bindings=None, skip=True),
        }
        for name, elem in elems.items():
            elem.ident = name
            if elem.skip:
                elem.validate.side_effect = elements.SkipBinding()
            elem.paths = {x: elems[x] for x in elem.t_paths}
            elem.bindings = (
                None if elem.t_bindings is None else elems[elem.t_bindings]
            )
        mocker.patch.object(
            controller.Controller, '_micropath_root', elems[None],
        )
        req = mocker.Mock(**{
            'path_info_peek.side_effect': ['a', '1', 'c', '2', ''],
            'urlvars': {},
        })
        inj = {}
        obj = controller.Controller()

        result = obj._micropath_resolve(req, inj)

        assert result == (elems['c'], True)
        url_vars = {
            'b': elems['b'].validate.return_value,
        }
        assert inj == url_vars
        assert req.urlvars == url_vars
        elems['b'].validate.assert_called_once_with(obj, inj, '1')
        elems['d'].validate.assert_called_once_with(obj, inj, '2')

    def test_micropath_resolve_unmatched(self, mocker):
        elems = {
            None: mocker.Mock(t_paths=['a'], t_bindings=None, skip=False),
            'a': mocker.Mock(t_paths=[], t_bindings='b', skip=False),
            'b': mocker.Mock(t_paths=['c'], t_bindings=None, skip=False),
            'c': mocker.Mock(t_paths=[], t_bindings=None, skip=False),
        }
        for name, elem in elems.items():
            elem.ident = name
            if elem.skip:
                elem.validate.side_effect = elements.SkipBinding()
            elem.paths = {x: elems[x] for x in elem.t_paths}
            elem.bindings = (
                None if elem.t_bindings is None else elems[elem.t_bindings]
            )
        mocker.patch.object(
            controller.Controller, '_micropath_root', elems[None],
        )
        req = mocker.Mock(**{
            'path_info_peek.side_effect': ['a', '1', 'c', '2', ''],
            'urlvars': {},
        })
        inj = {}
        obj = controller.Controller()

        result = obj._micropath_resolve(req, inj)

        assert result == (elems['c'], True)
        url_vars = {
            'b': elems['b'].validate.return_value,
        }
        assert inj == url_vars
        assert req.urlvars == url_vars
        elems['b'].validate.assert_called_once_with(obj, inj, '1')

    def test_micropath_delegation_base(self, mocker):
        req = mocker.Mock(method='GET')
        meth = mocker.Mock()
        elem = mocker.Mock(methods={'GET': meth})
        obj = controller.Controller()

        result = obj._micropath_delegation(req, elem)

        assert result == (meth.func, meth.delegation)

    def test_micropath_delegation_elem_delegation(self, mocker):
        req = mocker.Mock(method='GET')
        meth = mocker.Mock(delegation=None)
        elem = mocker.Mock(methods={'GET': meth})
        obj = controller.Controller()

        result = obj._micropath_delegation(req, elem)

        assert result == (meth.func, elem.delegation)

    def test_micropath_delegation_head(self, mocker):
        req = mocker.Mock(method='HEAD')
        meth = mocker.Mock()
        elem = mocker.Mock(methods={'GET': meth})
        obj = controller.Controller()

        result = obj._micropath_delegation(req, elem)

        assert result == (meth.func, meth.delegation)

    def test_micropath_delegation_none(self, mocker):
        req = mocker.Mock(method='GET')
        meth = mocker.Mock()
        elem = mocker.Mock(methods={None: meth})
        obj = controller.Controller()

        result = obj._micropath_delegation(req, elem)

        assert result == (meth.func, meth.delegation)

    def test_micropath_delegation_no_method(self, mocker):
        req = mocker.Mock(method='GET')
        elem = mocker.Mock(methods={})
        obj = controller.Controller()

        result = obj._micropath_delegation(req, elem)

        assert result == (None, elem.delegation)

    def test_micropath_methods_base(self, mocker):
        elem = mocker.Mock(methods={'POST': 'poster'})
        obj = controller.Controller()

        result = obj._micropath_methods(elem)

        assert result == set(['POST', 'OPTIONS'])

    def test_micropath_methods_with_get(self, mocker):
        elem = mocker.Mock(methods={'POST': 'poster', 'GET': 'getter'})
        obj = controller.Controller()

        result = obj._micropath_methods(elem)

        assert result == set(['POST', 'OPTIONS', 'GET', 'HEAD'])

    def test_micropath_methods_all(self, mocker):
        elem = mocker.Mock(methods={None: 'method'})
        obj = controller.Controller()

        result = obj._micropath_methods(elem)

        assert result == controller.Controller.micropath_methods

    def test_micropath_methods_all_extra(self, mocker):
        elem = mocker.Mock(methods={None: 'method', 'PATCH': 'patcher'})
        obj = controller.Controller()

        result = obj._micropath_methods(elem)

        assert result == (
            controller.Controller.micropath_methods | set(['PATCH'])
        )

    def test_micropath_construct(self, mocker):
        other = mocker.Mock()
        obj = controller.Controller()

        result = obj.micropath_construct(other, {'a': 1, 'b': 2, 'c': 3})

        assert result == other.return_value
        other.assert_called_once_with(a=1, b=2, c=3)

    def test_micropath_server_error_base(self, mocker):
        mock_format_exception = mocker.patch.object(
            controller.traceback, 'format_exception',
            return_value=['line1\n', 'line2\n', 'line3\n'],
        )
        mock_HTTPInternalServerError = mocker.patch.object(
            controller.webob.exc, 'HTTPInternalServerError',
        )
        obj = controller.Controller()

        result = obj.micropath_server_error('req', ('type', 'value', 'tb'))

        assert result == mock_HTTPInternalServerError.return_value
        mock_format_exception.assert_not_called()
        mock_HTTPInternalServerError.assert_called_once_with(None)

    def test_micropath_server_error_debug(self, mocker):
        mock_format_exception = mocker.patch.object(
            controller.traceback, 'format_exception',
            return_value=['line1\n', 'line2\n', 'line3\n'],
        )
        mock_HTTPInternalServerError = mocker.patch.object(
            controller.webob.exc, 'HTTPInternalServerError',
        )
        obj = controller.Controller()
        obj.micropath_debug = True

        result = obj.micropath_server_error('req', ('type', 'value', 'tb'))

        assert result == mock_HTTPInternalServerError.return_value
        mock_format_exception.assert_called_once_with('type', 'value', 'tb')
        mock_HTTPInternalServerError.assert_called_once_with(
            'line1\nline2\nline3\n',
        )

    def test_micropath_request_error_base(self, mocker):
        mock_format_exception = mocker.patch.object(
            controller.traceback, 'format_exception',
            return_value=['line1\n', 'line2\n', 'line3\n'],
        )
        mock_HTTPBadRequest = mocker.patch.object(
            controller.webob.exc, 'HTTPBadRequest',
        )
        obj = controller.Controller()

        result = obj.micropath_request_error(
            'req', 'key', ('type', 'value', 'tb'),
        )

        assert result == mock_HTTPBadRequest.return_value
        mock_format_exception.assert_not_called()
        mock_HTTPBadRequest.assert_called_once_with(None)

    def test_micropath_request_error_debug(self, mocker):
        mock_format_exception = mocker.patch.object(
            controller.traceback, 'format_exception',
            return_value=['line1\n', 'line2\n', 'line3\n'],
        )
        mock_HTTPBadRequest = mocker.patch.object(
            controller.webob.exc, 'HTTPBadRequest',
        )
        obj = controller.Controller()
        obj.micropath_debug = True

        result = obj.micropath_request_error(
            'req', 'key', ('type', 'value', 'tb'),
        )

        assert result == mock_HTTPBadRequest.return_value
        mock_format_exception.assert_called_once_with('type', 'value', 'tb')
        mock_HTTPBadRequest.assert_called_once_with(
            'Accessing request attribute "key":\nline1\nline2\nline3\n',
        )

    def test_micropath_not_found(self, mocker):
        mock_HTTPNotFound = mocker.patch.object(
            controller.webob.exc, 'HTTPNotFound',
        )
        obj = controller.Controller()

        result = obj.micropath_not_found('req', '/path/info')

        assert result == mock_HTTPNotFound.return_value
        mock_HTTPNotFound.assert_called_once_with()

    def test_micropath_not_implemented(self, mocker):
        mock_HTTPNotImplemented = mocker.patch.object(
            controller.webob.exc, 'HTTPNotImplemented',
        )
        obj = controller.Controller()

        result = obj.micropath_not_implemented('req', 'PUT')

        assert result == mock_HTTPNotImplemented.return_value
        mock_HTTPNotImplemented.assert_called_once_with()

    def test_micropath_options(self, mocker):
        mock_HTTPNoContent = mocker.patch.object(
            controller.webob.exc, 'HTTPNoContent',
        )
        obj = controller.Controller()

        result = obj.micropath_options('request', ['GET', 'HEAD', 'POST'])

        assert result == mock_HTTPNoContent.return_value
        mock_HTTPNoContent.assert_called_once_with(
            headers={'Allow': 'GET,HEAD,POST'},
        )

    def test_micropath_run_base(self, mocker):
        mock_make_server = mocker.patch.object(
            controller.simple_server, 'make_server',
        )
        server = mock_make_server.return_value
        obj = controller.Controller()

        obj.micropath_run()

        mock_make_server.assert_called_once_with('127.0.0.1', 8000, obj)
        server.serve_forever.assert_called_once_with()

    def test_micropath_run_alt(self, mocker):
        mock_make_server = mocker.patch.object(
            controller.simple_server, 'make_server',
        )
        server = mock_make_server.return_value
        obj = controller.Controller()

        obj.micropath_run('10.0.0.1', 5000)

        mock_make_server.assert_called_once_with('10.0.0.1', 5000, obj)
        server.serve_forever.assert_called_once_with()



# Copyright (C) 2018 by Kevin L. Mitchell <klmitch@mit.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import micropath

from tests.function import utils


class BookController(micropath.Controller):
    @micropath.route('get')
    def index(self, request, sub_id=None):
        return 'book::index(sub_id=%s)' % utils.safestr(sub_id)

    @micropath.route('post')
    def create(self, request, sub_id=None):
        return 'book::create(sub_id=%s)' % utils.safestr(sub_id)

    book_id = micropath.bind()

    @book_id.route('get')
    def get(self, request, book_id, sub_id=None):
        return 'book::get(book_id=%s, sub_id=%s)' % (
            utils.safestr(book_id), utils.safestr(sub_id),
        )

    @book_id.route('put')
    def update(self, request, book_id, sub_id=None):
        return 'book::update(book_id=%s, sub_id=%s)' % (
            utils.safestr(book_id), utils.safestr(sub_id),
        )

    @book_id.route('delete')
    def delete(self, request, book_id, sub_id=None):
        return 'book::delete(book_id=%s, sub_id=%s)' % (
            utils.safestr(book_id), utils.safestr(sub_id),
        )


class SubscriberController(micropath.Controller):
    @micropath.route('get')
    def index(self, request):
        return 'sub::index()'

    @micropath.route('post')
    def create(self, request):
        return 'sub::create()'

    sub_id = micropath.bind()

    @sub_id.route('get')
    def get(self, request, sub_id):
        return 'sub::get(sub_id=%s)' % utils.safestr(sub_id)

    @sub_id.route('put')
    def update(self, request, sub_id):
        return 'sub::update(sub_id=%s)' % utils.safestr(sub_id)

    @sub_id.route('delete')
    def delete(self, request, sub_id):
        return 'sub::delete(sub_id=%s)' % utils.safestr(sub_id)

    books = sub_id.path().mount(BookController)


class TestSimple(object):
    def test_subscriber_index(self):
        controller = SubscriberController()

        status, _headers, body = utils.invoke(
            controller, '/',
            method='GET',
        )

        assert status == '200 OK'
        assert body == b'sub::index()'

    def test_subscriber_index_url_for(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com')

        result = req.url_for(controller.index)

        assert result == 'http://example.com/'

    def test_subscriber_index_url_for_scriptname(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com/api')

        result = req.url_for(controller.index)

        assert result == 'http://example.com/api/'

    def test_subscriber_create(self):
        controller = SubscriberController()

        status, _headers, body = utils.invoke(
            controller, '/',
            method='POST',
        )

        assert status == '200 OK'
        assert body == b'sub::create()'

    def test_subscriber_create_url_for(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com')

        result = req.url_for(controller.create)

        assert result == 'http://example.com/'

    def test_subscriber_create_url_for_scriptname(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com/api')

        result = req.url_for(controller.create)

        assert result == 'http://example.com/api/'

    def test_subscriber_options(self):
        controller = SubscriberController()

        status, headers, body = utils.invoke(
            controller, '/',
            method='OPTIONS',
        )

        assert status == '204 No Content'
        assert 'allow' in headers
        assert headers['allow'] == 'GET,HEAD,OPTIONS,POST'
        assert body == b''

    def test_subscriber_other(self):
        controller = SubscriberController()

        status, _headers, _body = utils.invoke(
            controller, '/',
            method='OTHER',
        )

        assert status == '501 Not Implemented'

    def test_subscriber_subid_get(self):
        controller = SubscriberController()

        status, _headers, body = utils.invoke(
            controller, '/1234',
            method='GET',
        )

        assert status == '200 OK'
        assert body == b'sub::get(sub_id=1234)'

    def test_subscriber_get_url_for(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com')

        result = req.url_for(controller.get, sub_id='1234')

        assert result == 'http://example.com/1234'

    def test_subscriber_get_url_for_scriptname(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com/api')

        result = req.url_for(controller.get, sub_id='1234')

        assert result == 'http://example.com/api/1234'

    def test_subscriber_subid_update(self):
        controller = SubscriberController()

        status, _headers, body = utils.invoke(
            controller, '/1234',
            method='PUT',
        )

        assert status == '200 OK'
        assert body == b'sub::update(sub_id=1234)'

    def test_subscriber_update_url_for(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com')

        result = req.url_for(controller.update, sub_id='1234')

        assert result == 'http://example.com/1234'

    def test_subscriber_update_url_for_scriptname(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com/api')

        result = req.url_for(controller.update, sub_id='1234')

        assert result == 'http://example.com/api/1234'

    def test_subscriber_subid_delete(self):
        controller = SubscriberController()

        status, _headers, body = utils.invoke(
            controller, '/1234',
            method='DELETE',
        )

        assert status == '200 OK'
        assert body == b'sub::delete(sub_id=1234)'

    def test_subscriber_delete_url_for(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com')

        result = req.url_for(controller.delete, sub_id='1234')

        assert result == 'http://example.com/1234'

    def test_subscriber_delete_url_for_scriptname(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com/api')

        result = req.url_for(controller.delete, sub_id='1234')

        assert result == 'http://example.com/api/1234'

    def test_subscriber_subid_options(self):
        controller = SubscriberController()

        status, headers, body = utils.invoke(
            controller, '/1234',
            method='OPTIONS',
        )

        assert status == '204 No Content'
        assert 'allow' in headers
        assert headers['allow'] == 'DELETE,GET,HEAD,OPTIONS,PUT'
        assert body == b''

    def test_subscriber_subid_other(self):
        controller = SubscriberController()

        status, _headers, _body = utils.invoke(
            controller, '/1234',
            method='OTHER',
        )

        assert status == '501 Not Implemented'

    def test_subscriber_not_found(self):
        controller = SubscriberController()

        status, _headers, _body = utils.invoke(
            controller, '/1234/not_found',
            method='GET',
        )

        assert status == '404 Not Found'

    def test_book_index(self):
        controller = SubscriberController()

        status, _headers, body = utils.invoke(
            controller, '/1234/books',
            method='GET',
        )

        assert status == '200 OK'
        assert body == b'book::index(sub_id=1234)'

    def test_book_index_url_for(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com')

        result = req.url_for(controller.books.index, sub_id='1234')

        assert result == 'http://example.com/1234/books'

    def test_book_index_url_for_scriptname(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com/api')

        result = req.url_for(controller.books.index, sub_id='1234')

        assert result == 'http://example.com/api/1234/books'

    def test_book_create(self):
        controller = SubscriberController()

        status, _headers, body = utils.invoke(
            controller, '/1234/books',
            method='POST',
        )

        assert status == '200 OK'
        assert body == b'book::create(sub_id=1234)'

    def test_book_create_url_for(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com')

        result = req.url_for(controller.books.create, sub_id='1234')

        assert result == 'http://example.com/1234/books'

    def test_book_create_url_for_scriptname(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com/api')

        result = req.url_for(controller.books.create, sub_id='1234')

        assert result == 'http://example.com/api/1234/books'

    def test_book_options(self):
        controller = SubscriberController()

        status, headers, body = utils.invoke(
            controller, '/1234/books',
            method='OPTIONS',
        )

        assert status == '204 No Content'
        assert 'allow' in headers
        assert headers['allow'] == 'GET,HEAD,OPTIONS,POST'
        assert body == b''

    def test_book_other(self):
        controller = SubscriberController()

        status, _headers, _body = utils.invoke(
            controller, '/1234/books',
            method='OTHER',
        )

        assert status == '501 Not Implemented'

    def test_book_bookid_get(self):
        controller = SubscriberController()

        status, _headers, body = utils.invoke(
            controller, '/1234/books/5678',
            method='GET',
        )

        assert status == '200 OK'
        assert body == b'book::get(book_id=5678, sub_id=1234)'

    def test_book_bookid_get_url_for(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com')

        result = req.url_for(
            controller.books.get,
            sub_id='1234',
            book_id='5678',
        )

        assert result == 'http://example.com/1234/books/5678'

    def test_book_bookid_get_url_for_scriptname(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com/api')

        result = req.url_for(
            controller.books.get,
            sub_id='1234',
            book_id='5678',
        )

        assert result == 'http://example.com/api/1234/books/5678'

    def test_book_bookid_update(self):
        controller = SubscriberController()

        status, _headers, body = utils.invoke(
            controller, '/1234/books/5678',
            method='PUT',
        )

        assert status == '200 OK'
        assert body == b'book::update(book_id=5678, sub_id=1234)'

    def test_book_bookid_update_url_for(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com')

        result = req.url_for(
            controller.books.update,
            sub_id='1234',
            book_id='5678',
        )

        assert result == 'http://example.com/1234/books/5678'

    def test_book_bookid_update_url_for_scriptname(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com/api')

        result = req.url_for(
            controller.books.update,
            sub_id='1234',
            book_id='5678',
        )

        assert result == 'http://example.com/api/1234/books/5678'

    def test_book_bookid_delete(self):
        controller = SubscriberController()

        status, _headers, body = utils.invoke(
            controller, '/1234/books/5678',
            method='DELETE',
        )

        assert status == '200 OK'
        assert body == b'book::delete(book_id=5678, sub_id=1234)'

    def test_book_bookid_delete_url_for(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com')

        result = req.url_for(
            controller.books.delete,
            sub_id='1234',
            book_id='5678',
        )

        assert result == 'http://example.com/1234/books/5678'

    def test_book_bookid_delete_url_for_scriptname(self):
        controller = SubscriberController()
        req = micropath.Request.blank('/', base_url='http://example.com/api')

        result = req.url_for(
            controller.books.delete,
            sub_id='1234',
            book_id='5678',
        )

        assert result == 'http://example.com/api/1234/books/5678'

    def test_book_bookid_options(self):
        controller = SubscriberController()

        status, headers, body = utils.invoke(
            controller, '/1234/books/5678',
            method='OPTIONS',
        )

        assert status == '204 No Content'
        assert 'allow' in headers
        assert headers['allow'] == 'DELETE,GET,HEAD,OPTIONS,PUT'
        assert body == b''

    def test_book_bookid_other(self):
        controller = SubscriberController()

        status, _headers, _body = utils.invoke(
            controller, '/1234/books/5678',
            method='OTHER',
        )

        assert status == '501 Not Implemented'



# Copyright (C) 2018 by Kevin L. Mitchell <klmitch@mit.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import collections
import functools
import inspect

import six


class WantSignature(object):
    """
    The signature of desired arguments for a function.  This collects
    the sets of keyword arguments that the function wants, and allows
    for the function to be called with only those arguments.  This is
    the core of the dependency injection implementation.
    """

    if six.PY2:
        @staticmethod
        def _getsig(func):
            """
            Obtain a function signature, for Python 2.  This analyzes the
            function argument specification and constructs the
            argument order, required set of arguments, and optional
            set of arguments.

            :param func: The function to analyze.

            :returns: A tuple of five elements.  The first element is
                      a list providing the order of positional
                      arguments.  The second element is a set of the
                      arguments that the function requires be
                      provided.  The third element is a set of the
                      arguments that the function wants, if they are
                      available.  The last two elements are booleans
                      indicating whether the function wants all
                      positional or keyword arguments, respectively.
            """

            # Get the argument specification
            argspec = inspect.getargspec(func)
            order = argspec.args[:]

            # Figure out how many are optional
            if argspec.defaults:
                defcnt = -len(argspec.defaults)
                required = set(order[:defcnt])
                optional = set(order[defcnt:])
            else:
                required = set(order)
                optional = set()

            return (
                order, required, optional,
                argspec.varargs is not None,
                argspec.keywords is not None,
            )
    else:
        @staticmethod
        def _getsig(func):
            """
            Obtain a function signature, for Python 3.  This analyzes the
            function argument specification and constructs the
            argument order, required set of arguments, and optional
            set of arguments.

            :param func: The function to analyze.

            :returns: A tuple of five elements.  The first element is
                      a list providing the order of positional
                      arguments.  The second element is a set of the
                      arguments that the function requires be
                      provided.  The third element is a set of the
                      arguments that the function wants, if they are
                      available.  The last two elements are booleans
                      indicating whether the function wants all
                      positional or keyword arguments, respectively.
            """

            # Initialize the data to return
            order = []
            required = set()
            optional = set()
            all_pos = False
            all_kw = False

            # Get the function signature
            sig = inspect.signature(func, follow_wrapped=False)

            # Process the function signature
            for param in sig.parameters.values():
                # Pick only positional-capable arguments for order
                if param.kind in (
                        param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD,
                ):
                    order.append(param.name)

                # Pick only keyword-capable arguments for required and
                # optional
                if param.kind in (
                        param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY,
                ):
                    # Presence of a default controls whether it's
                    # required or optional
                    if param.default is param.empty:
                        required.add(param.name)
                    else:
                        optional.add(param.name)

                # Toggle all_pos/all_kw as needed
                if param.kind == param.VAR_POSITIONAL:
                    all_pos = True
                elif param.kind == param.VAR_KEYWORD:
                    all_kw = True

            return (order, required, optional, all_pos, all_kw)

    @classmethod
    def from_func(cls, func, wrapped=None, provides=None,
                  required=None, optional=None):
        """
        Given a function, retrieve its signature.  This will additionally
        cache the signature on the function for future calls.

        :param func: The function to retrieve the signature for.
        :param wrapped: The function being wrapped by ``func``.  This
                        is used if ``func`` is to be a wrapper for
                        another function.
        :param provides: An iterable of function arguments that the
                         wrapper provides to the wrapped function.
                         Only meaningful if ``wrapped`` is provided.
        :param required: An iterable of function arguments that the
                         function requires.  Only meaningful if
                         ``func`` has an all-keywords argument; in
                         this case, the specified arguments are added
                         to the "required" set, and the ``all_kw``
                         value is set to ``False``.
        :param optional: An iterable of function arguments that the
                         function wants if they are available.  Only
                         meaningful if ``func`` has an all-keywords
                         argument; in this case, the specified
                         arguments are added to the "optional" set,
                         and the ``all_kw`` value is set to ``False``.

        :returns: The signature of the specified function.
        :rtype: ``WantSignature``
        """

        # Construct and set the signature
        if not hasattr(func, '_micropath_signature'):
            # First, collect the function signature
            order, func_req, func_opt, all_pos, all_kw = cls._getsig(func)

            # If it's wrapping something, deal with that
            if wrapped is not None:
                wrapped_sig = cls.from_func(wrapped)

                # First, merge in the required and optional
                func_opt = (
                    (func_opt - wrapped_sig.required) |
                    (wrapped_sig.optional - func_req)
                )
                func_req |= wrapped_sig.required

                # Next, discard provided arguments
                provided = set(provides or [])
                func_req -= provided
                func_opt -= provided

            # Handle required and optional
            if all_kw:
                if required is not None:
                    func_req |= set(required)
                    all_kw = False
                if optional is not None:
                    func_opt |= set(optional)
                    all_kw = False

            # Set the function signature
            func._micropath_signature = cls(
                func, order, func_req, func_opt, all_pos, all_kw,
            )

        return func._micropath_signature

    def __init__(self, func, arg_order, required, optional, all_pos, all_kw):
        """
        Initialize a ``WantSignature`` instance.

        :param func: The function the signature is for.  When the
                     signature is invoked, the function will be
                     invoked.
        :param arg_order: A list of the names of arguments, in the
                          order they are defined.
        :type arg_order: ``list`` of ``str``
        :param required: A set of the argument names that must be
                         provided.  Attempts to call the function
                         without one of these arguments will result in
                         an error.
        :type required: ``set`` of ``str``
        :param optional: A set of the argument names that may
                         optionally be provided.  These will typically
                         be arguments for which default values are
                         set.
        :type optional: ``set`` of ``str``
        :param bool all_pos: If ``True``, the function will accept all
                             positional arguments.
        :param bool all_kw: If ``True``, the function will accept all
                            keyword arguments.
        """

        # Make sure there's no overlap
        if required & optional:
            raise ValueError('overlap between required and optional')

        self.func = func
        self.arg_order = arg_order
        self.required = required
        self.optional = optional
        self.all_pos = all_pos
        self.all_kw = all_kw
        self.all_args = required | optional

    def __contains__(self, keyword):
        """
        Used to determine if the function wants a particular keyword
        argument.

        :param str keyword: The name of the keyword argument.

        :returns: A ``True`` value if the signature indicates the
                  function wants the keyword argument (as either
                  required or optional), or ``False`` otherwise.
        :rtype: ``bool``
        """

        return self.all_kw or keyword in self.all_args

    def __call__(self, args, kwargs, additional=None):
        """
        Call the function, passing it the desired arguments and keyword
        arguments.

        :param tuple args: A tuple of positional arguments.  In
                           addition to being passed to the function,
                           this will also skip collecting a value for
                           these arguments from the provided keyword
                           arguments.
        :param dict kwargs: The keyword arguments available to pass to
                            the function.  Only those the function
                            wants, as defined by the signature, will
                            actually be passed.
        :param dict additional: An optional dictionary of additional
                                keyword arguments.  The values in this
                                dictionary will override those
                                provided by ``kwargs``.

        :returns: The return value of calling the function with the
                  specified positional and keyword arguments.
        """

        # Make sure there aren't too many positional arguments
        if not self.all_pos and len(args) > len(self.arg_order):
            raise TypeError(
                'too many positional arguments: got %d, can handle at '
                'most %d' %
                (len(args), len(self.arg_order))
            )

        # Canonicalize additional
        if not additional:
            additional = {}

        # Construct the set of desired keyword arguments
        desired = self.all_args
        if self.all_kw:
            desired |= set(additional) | set(kwargs)
        satisfied = set(self.arg_order[:len(args)])
        desired -= satisfied

        # Construct the keyword arguments
        real_kw = {
            key: additional[key] if key in additional else kwargs[key]
            for key in desired if key in additional or key in kwargs
        }

        # Make sure we got all the required arguments
        missing = self.required - set(real_kw) - satisfied
        if missing:
            raise TypeError(
                'missing required keyword arguments: "%s"' %
                '", "'.join(sorted(missing))
            )

        # Call the function
        return self.func(*args, **real_kw)


class InjectorCleanup(object):
    """
    A context manager for cleaning up keys added to an ``Injector``.
    This is used to mitigate reference loops that could cause problems
    for the Python garbage collector.
    """

    def __init__(self, injector):
        """
        Initialize an ``InjectorCleanup`` instance.

        :param injector: The injector to clean up.
        :type injector: ``Injector``
        """

        self.injector = injector
        self.keep = None

    def __enter__(self):
        """
        Enter the context manager.  This copies the set of existing keys
        for later cleanup operations.

        :returns: The injector.
        :rtype: ``Injector``
        """

        # Must be None
        assert self.keep is None

        # Copy the current set of keys
        self.keep = set(self.injector._keys)

        return self.injector

    def __exit__(self, exc_type, exc_value, exc_tb):
        """
        Exit the context manager.  This deletes all keys added to the
        injector since the context manager was entered.  The
        exception, if any, is not handled.

        :param exc_type: The type of the exception that was raised, or
                         ``None``.
        :param exc_value: The exception that was raised, or ``None``.
        :param exc_tb: The traceback of the exception that was raised,
                       or ``None``.

        :returns: A ``None`` value, to indicate that the exception was
                  not handled.
        """

        # Must not be None
        assert self.keep is not None

        # Delete the added keys
        for key in self.injector._keys - self.keep:
            del self.injector[key]

        # Reset keep
        self.keep = None

        return None


class Injector(collections.MutableMapping):
    """
    A mutable mapping that provides the collection of arguments that
    are available to be passed to a function.  This is used in
    conjunction with the ``WantSignature`` class for dependency
    injection.
    """

    def __init__(self):
        """
        Initialize an ``Injector`` instance.
        """

        # Set up the maps and the key set
        self._available = {}  # values
        self._deferred = {}  # callables generating values
        self._keys = set()  # efficiency enhancement containing keys

    def __len__(self):
        """
        Return the number of items in the mapping.

        :returns: The number of items in the mapping.
        :rtype: ``int``
        """

        return len(self._keys)

    def __iter__(self):
        """
        Iterate over the keys in the mapping.

        :returns: An iterator over the keys in the mapping.
        """

        return iter(self._keys)

    def __getitem__(self, key):
        """
        Retrieve the value of an item.  If the value has not been set, but
        a deferred callable for it has, that deferred callable will be
        called and the value set, prior to returning it.

        :param key: The key of the item to retrieve.

        :returns: The value associated with that key.
        """

        # Handle the KeyError case first
        if key not in self._keys:
            raise KeyError(key)

        # If it's not available, we'll need to call the deferred
        # action
        if key not in self._available:
            self._available[key] = self(self._deferred[key])

        return self._available[key]

    def __setitem__(self, key, value):
        """
        Set the value of an item.  This will mask any deferred callable
        that has been set for the same key.

        :param key: The key of the item to set.
        :param value: The value to associate with the key.
        """

        self._available[key] = value
        self._keys.add(key)

    def __delitem__(self, key):
        """
        Delete the value of an item.  This also discards any deferred
        callable that has been set for the key.

        :param key: The key of the item to delete.
        """

        # Handle the KeyError case first
        if key not in self._keys:
            raise KeyError(key)

        # Pop it off
        self._available.pop(key, None)
        self._deferred.pop(key, None)
        self._keys.discard(key)

    def __call__(self, *args, **kwargs):
        """
        Call a function, injecting the keyword arguments desired by the
        function from the keys defined in this mapping.  The first
        positional parameter must be the function to invoke.
        Additional positional arguments are passed directly to the
        function, and additional keyword arguments override values
        from this mapping.  This is the entrypoint for dependency
        injection.
        """

        if len(args) < 1:
            raise TypeError('call requires at least one positional argument')

        # Split the function and arguments
        func = args[0]
        args = args[1:]

        # Unwrap class and instance methods
        if inspect.ismethod(func):
            obj = six.get_method_self(func)
            func = six.get_method_function(func)

            # Update the args
            args = (obj,) + args

        # Get the function's injection signature
        sig = WantSignature.from_func(func)

        # Call the function
        return sig(args, self, kwargs)

    def set_deferred(self, key, func):
        """
        Set a deferred callable for a key.  This may be used when a value
        for a key should only be generated if the function being
        called actually wants the value, usually because generating
        the value is somewhat expensive.

        :param key: The key the value should be associated with.
        :param func: A callable that will generate the value to
                     associate with the key.
        """

        self._deferred[key] = func

    def cleanup(self):
        """
        Returns a context manager that ensures that keys added during the
        execution of the ``with`` statement are deleted at the end.
        This is used to help mitigate reference loops by explicitly
        breaking them.

        :returns: A context manager for cleaning up the injector.
        :rtype: ``InjectorCleanup``
        """

        return InjectorCleanup(self)


def inject(required=None, optional=None):
    """
    A function decorator that allows dependency injection to be
    tailored.  In most cases, it is not necessary to use this
    decorator; it may be used when a function takes all keyword
    arguments--i.e., a ``**kwargs`` or similar argument is
    present--but the developer wishes to restrict the set of
    injectible arguments.

    :param required: An iterable of function arguments that the
                     function requires.
    :param optional: An iterable of function arguments that the
                     function wants if they are available.

    :returns: A function decorator that sets the dependency injection
              metadata of the function as desired.
    """

    # The actual decorator; just calls from_func() with appropriate
    # arguments
    def decorator(func):
        WantSignature.from_func(func, required=required, optional=optional)
        return func

    return decorator


# Sentinel to indicate required/optional not explicitly provided
_unset = object()


def wraps(wrapped, assigned=functools.WRAPPER_ASSIGNMENTS,
          updated=functools.WRAPPER_UPDATES, provides=None,
          required=_unset, optional=_unset):
    """
    A function decorator for wrapping decorators.  This works just
    like ``six.wraps()`` (which in turn works just like
    ``functools.wraps()``), but additionally manages dependency
    injection metadata, allowing decorators to request data
    independent of the function they wrap.

    :param wrapped: The function that is being wrapped.
    :param assigned: A tuple naming the attributes assigned directly
                     from the wrapped function to the wrapper
                     function.  Defaults to
                     ``functools.WRAPPER_ASSIGNMENTS``.
    :param updated: A tuple naming the attributes of the wrapper that
                    are updated with the corresponding attribute from
                    the wrapped function.  Defaults to
                    ``functools.WRAPPER_UPDATES``.
    :param provides: An iterable of function arguments that the
                     wrapper provides to the wrapped function.
    :param required: An iterable of function arguments that the
                     function requires.  Only meaningful if ``func``
                     has an all-keywords argument; in this case, the
                     specified arguments are added to the "required"
                     set, and the ``all_kw`` value is set to
                     ``False``.  Note that by default, ``required`` is
                     set to the empty list; pass ``None`` for both
                     ``required`` and ``optional`` to pass all
                     injectible keyword arguments to the wrapper.
    :param optional: An iterable of function arguments that the
                     function wants if they are available.  Only
                     meaningful if ``func`` has an all-keywords
                     argument; in this case, the specified arguments
                     are added to the "optional" set, and the
                     ``all_kw`` value is set to ``False``.  Note that
                     by default, ``optional`` is set to the empty
                     list; pass ``None`` for both ``required`` and
                     ``optional`` to pass all injectible keyword
                     arguments to the wrapper.

    :returns: A function decorator that properly updates the
              attributes of the wrapper function, while also setting
              the dependency injection metadata of the wrapper
              function as desired.
    """

    # The actual decorator
    def decorator(func):
        # Generate the signature first
        sig = WantSignature.from_func(
            func, wrapped, provides,
            [] if required is _unset else required,
            [] if optional is _unset else optional,
        )

        # Next, wrap it
        func = six.wraps(wrapped, assigned, updated)(func)

        # The wrapper may override the signature, so reset it
        func._micropath_signature = sig

        return func

    return decorator


def call_wrapped(func, args, kwargs):
    """
    Call a wrapped function with appropriate dependency injection.
    This is for use by decorators that wrap a function that will be
    called via dependency injection, and ensures that the function is
    called only with the desired keyword arguments.

    :param func: The wrapped function.
    :param args: A tuple of positional arguments to pass to the
                 function.
    :param kwargs: A dictionary of keyword arguments to pass to the
                   function.  If the decorator provides arguments
                   (uses the ``micropath.wraps()`` decorator with the
                   ``provides`` argument), those argument values must
                   be present in ``kwargs``.

    :returns: The result of calling the wrapped function.
    """

    # Get the function's injection signature
    sig = WantSignature.from_func(func)

    # Call the function
    return sig(args, kwargs)


def wants(func, keyword):
    """
    Determine if a function wants a particular keyword argument.

    :param func: The function.
    :param str keyword: The keyword argument.

    :returns: A ``True`` value if the function wants the keyword
              argument, or ``False`` otherwise.
    :rtype: ``bool``
    """

    # Get the function's injection signature
    sig = WantSignature.from_func(func)

    # See if it wants the argument
    return keyword in sig



#!/usr/bin/env python

import setuptools


# Utility function to read the README file
def readfile(filename):
    with open(filename) as f:
        return f.read()


# Utility function to read requirements.txt files
def readreq(filename):
    result = []
    with open(filename) as f:
        for line in f:
            line = line.strip()

            # Process requirement file references
            if line.startswith('-r '):
                subfilename = line.split(None, 1)[-1].split('#', 1)[0].strip()
                if subfilename:
                    result += readreq(subfilename)
                continue

            # Strip out "-e" prefixes
            if line.startswith('-e '):
                line = line.split(None, 1)[-1]

            # Detect URLs in the line
            idx = line.find('#egg=')
            if idx >= 0:
                line = line[idx + 5:]

            # Strip off any comments
            line = line.split('#', 1)[0].strip()

            # Save the requirement
            if line:
                result.append(line.split('#', 1)[0].strip())

    return result


setuptools.setup(
    name='micropath',
    version='0.1.0',
    author='Kevin L. Mitchell',
    author_email='klmitch@mit.edu',
    url='https://github.com/klmitch/micropath',
    description="The micropath Web Application Framework",
    long_description=readfile('README.rst'),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: No Input/Output (Daemon)',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
    ],
    packages=setuptools.find_packages(exclude=['tests', 'tests.*']),
    install_requires=readreq('requirements.txt'),
    tests_require=readreq('test-requirements.txt'),
)



# Copyright (C) 2018 by Kevin L. Mitchell <klmitch@mit.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

import abc

import six

from micropath import injector


@six.add_metaclass(abc.ABCMeta)
class Element(object):
    """
    Represent an abstract path element.  This can either be a constant
    component (``Path``); a variable component (``Binding``); or an
    HTTP method (``Method``).
    """

    def __init__(self, ident, parent=None):
        """
        Initialize an ``Element`` instance.

        :param str ident: The identifier for the component.  For
                          constant components, this will be the
                          expected value of the component; for
                          variable components, this will be the name
                          of the variable to contain the value.
        :param parent: The parent element.  Defaults to ``None``,
                       indicating the root of the element tree.
        :type parent: ``Element``
        """

        # Save the parameter
        self.ident = ident
        self.parent = parent

        # Set up subordinate lists
        self.paths = {}
        self.bindings = None
        self.methods = {}

        # For delegation to other controllers
        self.delegation = None

    @abc.abstractmethod
    def set_ident(self, ident):
        """
        Set the element identifier.  This can be used in the case that the
        identifier was not set at construction time.  Subclasses must
        call the superclass abstract method to set the identifier,
        then must update the subordinate lists in the parent object,
        if a parent exists.

        :param str ident: The identifier for the component.  For
                          constant components, this will be the
                          expected value of the component; for
                          variable components, this will be the name
                          of the variable to contain the value.

        :raises ValueError:
            The element identifier has already been set.
        """

        # Elements must be immutable
        if self.ident:
            raise ValueError('ident has already been set to "%s"' % self.ident)

        # Set the identifier
        self.ident = ident

    def path(self, ident=None):
        """
        Construct a new subordinate ``Path`` element.

        :param str ident: The identifier for the component.  For
                          constant components, this will be the
                          expected value of the component.  May be
                          ``None`` (the default) to indicate that the
                          metaclass should set the value from the name
                          of the class variable to which the element
                          is assigned.

        :returns: The newly constructed element.
        :rtype: ``Path``
        """

        # Construct the new path
        elem = Path(ident, parent=self)

        # If it has an identifier, add it to the lists
        if elem.ident:
            if elem.ident in self.paths:
                raise ValueError(
                    'Path element for "%s" already exists' % elem.ident,
                )
            self.paths[elem.ident] = elem

        return elem

    def bind(self, ident=None):
        """
        Construct a new subordinate ``Binding`` element.

        :param str ident: The identifier for the component.  For
                          variable components, this will be the name
                          of the variable to contain the value.  May
                          be ``None`` (the default) to indicate that
                          the metaclass should set the value from the
                          name of the class variable to which the
                          element is assigned.

        :returns: The newly constructed element.
        :rtype: ``Binding``
        """

        # Construct the new binding
        elem = Binding(ident, parent=self)

        # If it has an identifier, add it to the lists
        if elem.ident:
            if self.bindings is not None:
                raise ValueError(
                    'Binding element for "%s" already exists' % elem.ident,
                )
            self.bindings = elem

        return elem

    def route(self, *methods):
        """
        Method decorator to route HTTP methods to a method.  Called with
        an optional list of method strings (which will be
        canonicalized to uppercase).  If no methods are specified, all
        otherwise undefined methods will be routed to the decorated
        method.  This decorator is designed to work both with and
        without arguments--that is, ``@obj.route`` is equivalent to
        ``@obj.route()``.

        :returns: Either the appropriately decorated method, or a
                  decorator for a method.
        """

        def decorator(func):
            # Construct the new Method objects
            if methods:
                for meth_str in methods:
                    if meth_str in self.methods:
                        continue
                    meth = Method(meth_str, func, parent=self)
                    self.methods[meth.ident] = meth
            else:
                meth = Method(None, func, parent=self)
                self.methods[meth.ident] = meth

            # Mark the function as a handler and save its element
            func._micropath_handler = True
            func._micropath_elem = self

            # Pre-compute its want signature
            injector.WantSignature.from_func(func)

            return func

        # Check if methods consists of a single callable element
        if len(methods) == 1 and callable(methods[0]):
            func = methods[0]
            methods = ()
            return decorator(func)

        # Check for duplicate idents
        dups = [meth_str for meth_str in methods if meth_str in self.methods]
        if dups:
            raise ValueError(
                'Method element(s) for "%s" already exist(s)' %
                '", "'.join(sorted(dups)),
            )

        return decorator

    def mount(self, delegation, *methods, **kwargs):
        """
        Mount another controller class at this path element.  Only one
        delegation is allowed.  The delegation can be restricted to a
        set of HTTP methods by listing the method name strings after
        the controller class; the default is to delegate all HTTP
        methods to the other controller.

        Additional keyword arguments will be passed when creating a
        ``Delegation``, which will subsequently pass them to the
        controller's ``micropath_construct()`` method.

        :param delegation: Another controller to delegate to.  This
                           may be a ``micropath.Controller`` class
                           (not instance), or an instance of a
                           specialized subclass of ``Delegation``; the
                           latter possibility enables mounting other
                           WSGI applications, among other uses.

        :returns: The delegation to be mounted at this path element.
        :rtype: ``Delegation``

        :raises ValueError:
            The delegation has already been set.
        """

        # Make sure the delegation hasn't already been set
        if self.delegation:
            raise ValueError('delegation has already been set')

        # Wrap the delegation, if necessary
        if not isinstance(delegation, Delegation):
            delegation = Delegation(delegation, kwargs)

        # Save ourselves into the delegation so the controller
        # metaclass can do its thing
        delegation.element = self

        # Set the delegation
        if methods:
            # Check for duplicate idents
            dups = [
                meth_str for meth_str in methods if meth_str in self.methods
            ]
            if dups:
                raise ValueError(
                    'Method element(s) for "%s" already exist(s)' %
                    '", "'.join(sorted(dups)),
                )

            # Method restrictions specified, so apply them
            for meth_str in methods:
                if meth_str in self.methods:
                    continue
                meth = Method(meth_str, None, parent=self)
                self.methods[meth.ident] = meth
                meth.delegation = delegation
        else:
            # Delegation on us
            self.delegation = delegation

        return delegation


class Root(Element):
    """
    Represent a root element.  The root element is like a ``Path``
    element, but has no identifier.  A ``micropath.Controller`` class
    has exactly one ``Root`` instance associated with it.
    """

    def __init__(self):
        """
        Initialize a ``Root`` instance.
        """

        # Initialize the superclass
        super(Root, self).__init__(None)

    def set_ident(self, ident):
        """
        Set the element identifier.  This can be used in the case that the
        identifier was not set at construction time.  Subclasses must
        call the superclass abstract method to set the identifier,
        then must update the subordinate lists in the parent object,
        if a parent exists.

        :param str ident: The identifier for the component.  For
                          constant components, this will be the
                          expected value of the component; for
                          variable components, this will be the name
                          of the variable to contain the value.

        :raises ValueError:
            The element identifier has already been set.
        """

        raise ValueError('ident has already been set')

    def add_elem(self, elem, ident=None):
        """
        Add an element to the root.

        :param elem: The element to add.
        :type elem: ``Element``
        :param str ident: The identifier for the component.  Optional.
                          If specified, the first ancestor element
                          without an identifier will have its
                          identifier set to this value.

        :raises ValueError:
            An invalid element is being added.  Could indicate an
            ancestor is missing an identifier (and ``ident`` was not
            set or has been consumed by another element), or an
            element is of an unknown class.
        """

        # Handle the possibility of loops
        seen = set()

        # Walk up the tree
        while True:
            # Loop protection
            assert id(elem) not in seen
            seen.add(id(elem))

            # Sanity-check the element
            if elem is self:
                # Guess it's already been added to us
                return
            elif isinstance(elem, Root):
                raise ValueError('Cannot add a Root element to a Root element')
            elif not isinstance(elem, Method) and not elem.ident:
                # Set the element's ident
                if ident:
                    # It can now be set
                    elem.set_ident(ident)
                    ident = None

            # Have we found the element to be added to the root?
            if not elem.parent:
                break

            # Walk up to the parent
            elem = elem.parent

        # We've found the element to be added to the root, so do so
        if isinstance(elem, Path):
            if elem.ident:
                if elem.ident in self.paths:
                    raise ValueError(
                        'Path element for "%s" already exists' % elem.ident,
                    )
                self.paths[elem.ident] = elem
        elif isinstance(elem, Binding):
            if elem.ident:
                if self.bindings is not None:
                    raise ValueError(
                        'Binding element for "%s" already exists' % elem.ident,
                    )
                self.bindings = elem
        elif isinstance(elem, Method):
            if elem.ident in self.methods:
                raise ValueError(
                    'Method element for "%s" already exists' % elem.ident,
                )
            self.methods[elem.ident] = elem
        else:
            raise ValueError(
                'encountered unknown element class "%s.%s"' %
                (elem.__class__.__module__, elem.__class__.__name__)
            )

        # Set the element's parent
        elem.parent = self


class Path(Element):
    """
    Represent a constant path element.
    """

    def set_ident(self, ident):
        """
        Set the element identifier.  This can be used in the case that the
        identifier was not set at construction time.  Subclasses must
        call the superclass abstract method to set the identifier,
        then must update the subordinate lists in the parent object,
        if a parent exists.

        :param str ident: The identifier for the component.  For
                          constant components, this will be the
                          expected value of the component; for
                          variable components, this will be the name
                          of the variable to contain the value.

        :raises ValueError:
            The element identifier has already been set.
        """

        super(Path, self).set_ident(ident)

        if self.parent:
            if self.ident in self.parent.paths:
                raise ValueError(
                    'Path element for "%s" already exists' % self.ident,
                )
            self.parent.paths[self.ident] = self


class SkipBinding(Exception):
    """
    An exception that may be raised by the validator function of a
    ``Binding``.  This special exception indicates that this binding
    does not match the component and the next component should be
    checked.
    """

    pass


class Binding(Element):
    """
    Represent a variable path element.  In addition to being in the
    tree, variable path elements have an optional validator function
    (use the ``@Binding.validator`` decorator to set).
    """

    def __init__(self, ident, parent=None):
        """
        Initialize a ``Binding`` instance.

        :param str ident: The identifier for the component.  For
                          variable components, this will be the name
                          of the variable to contain the value.
        :param parent: The parent element.  Defaults to ``None``,
                       indicating the root of the element tree.
        :type parent: ``Element``
        """

        # Call the superclass first
        super(Binding, self).__init__(ident, parent)

        # Initialize the validator and formatter
        self._validator = None
        self._formatter = None

    def set_ident(self, ident):
        """
        Set the element identifier.  This can be used in the case that the
        identifier was not set at construction time.  Subclasses must
        call the superclass abstract method to set the identifier,
        then must update the subordinate lists in the parent object,
        if a parent exists.

        :param str ident: The identifier for the component.  For
                          constant components, this will be the
                          expected value of the component; for
                          variable components, this will be the name
                          of the variable to contain the value.

        :raises ValueError:
            The element identifier has already been set.
        """

        super(Binding, self).set_ident(ident)

        if self.parent:
            if self.parent.bindings is not None:
                raise ValueError(
                    'Binding element for "%s" already exists' % self.ident,
                )
            self.parent.bindings = self

    def validator(self, func):
        """
        A function decorator that sets the validator function for the
        binding.  If this decorator is not used, no validator function
        is set and the binding always validates.

        The validator function must define a ``value`` parameter,
        which will receive the text value of the path element.
        Dependency injection is used, so the validator function may
        request any other request parameter.  Note that the return
        value of the validator function becomes the value of the
        variable.  Also note that the validator function may raise
        ``SkipBinding`` to cause the next binding to be evaluated; any
        other exception will be bubbled up to the client.

        :param func: The function being decorated.

        :returns: The function that was decorated (``func``).

        :raises ValueError:
            The validator function has already been set.
        """

        # Elements must be immutable
        if self._validator:
            raise ValueError('validator has already been set')

        # Save the validator
        self._validator = func

        # Pre-compute its want signature
        injector.WantSignature.from_func(func)

        # Return the function
        return func

    def validate(self, obj, inj, value):
        """
        Validate a value.  This invokes the validator, if one is set, and
        returns the suitably transformed value to assign to the
        variable.

        :param obj: The instance of the controller class.  This must
                    be passed so that ``self`` can be present in the
                    decorated method.
        :param inj: The dependency injector.
        :type inj: ``micropath.injector.Injector``
        :param value: The value to validate.

        :returns: A suitably transformed value.

        :raises SkipBinding:
            This binding is not applicable to the specified value; try
            any other defined bindings.
        """

        # Call the validator
        if self._validator:
            return inj(self._validator, obj, value=value)

        return value

    def formatter(self, func):
        """
        A function decorator that sets the formatter function for the
        binding.  If this decorator is not used, binding values will
        be converted to URL elements using a simple string conversion
        (``six.text_type`` called on the value).

        Besides ``self``, the formatter function is only passed the
        value; unlike with the ``@validator`` function, dependency
        injection is not available.  The formatter function must
        return a string value.

        :param func: The function being decorated.

        :returns: The function that was decorated (``func``).

        :raises ValueError:
            The formatter function has already been set.
        """

        # Elements must be immutable
        if self._formatter:
            raise ValueError('formatter has already been set')

        # Save the formatter and return the decorated function
        self._formatter = func
        return func

    def format(self, obj, value):
        """
        Format a value.  This converts the value of a binding parameter
        back into a textual URL path component.  If a ``@formatter``
        function has not been set, the value is converted using simple
        string conversion (``six.text_type`` called on the value).

        :param obj: The instance of the controller class.  This must
                    be passed so that ``self`` can be present in the
                    decorated method.
        :param value: The value to validate.

        :returns: A suitably transformed value.
        :rtype: ``str``
        """

        # Call the formatter
        if self._formatter:
            return self._formatter(obj, value)

        return six.text_type(value)


class Method(Element):
    """
    Represent a method route.
    """

    def __init__(self, ident, func, parent=None):
        """
        Initialize a ``Method`` instance.

        :param str ident: The HTTP method.  This will be canonicalized
                          to an uppercase string.  Note that a
                          ``None`` value indicates a fallback for
                          otherwise unspecified methods.
        :param func: The callable that will handle the method.
        :param parent: The parent element.  Defaults to ``None``,
                       indicating the root of the element tree.
        :type parent: ``Element``
        """

        # Canonicalize the ident
        if isinstance(ident, six.string_types):
            ident = ident.upper()

        # Initialize the superclass
        super(Method, self).__init__(ident, parent)

        # Save the function
        self.func = func

    def set_ident(self, ident):
        """
        Set the element identifier.  This can be used in the case that the
        identifier was not set at construction time.  Subclasses must
        call the superclass abstract method to set the identifier,
        then must update the subordinate lists in the parent object,
        if a parent exists.

        :param str ident: The identifier for the component.  For
                          constant components, this will be the
                          expected value of the component; for
                          variable components, this will be the name
                          of the variable to contain the value.

        :raises ValueError:
            The element identifier has already been set.
        """

        raise ValueError('ident has already been set')

    def path(self, ident=None):
        """
        Construct a new subordinate ``Path`` element.

        :param str ident: The identifier for the component.  For
                          constant components, this will be the
                          expected value of the component.  May be
                          ``None`` (the default) to indicate that the
                          metaclass should set the value from the name
                          of the class variable to which the element
                          is assigned.

        :returns: The newly constructed element.
        :rtype: ``Path``
        """

        raise ValueError('cannot attach a path to a method')

    def bind(self, ident=None):
        """
        Construct a new subordinate ``Binding`` element.

        :param str ident: The identifier for the component.  For
                          variable components, this will be the name
                          of the variable to contain the value.  May
                          be ``None`` (the default) to indicate that
                          the metaclass should set the value from the
                          name of the class variable to which the
                          element is assigned.

        :returns: The newly constructed element.
        :rtype: ``Binding``
        """

        raise ValueError('cannot attach a binding to a method')

    def route(self, *methods):
        """
        Method decorator to route HTTP methods to a method.  Called with
        an optional list of method strings (which will be
        canonicalized to uppercase).  If no methods are specified, all
        otherwise undefined methods will be routed to the decorated
        method.  This decorator is designed to work both with and
        without arguments--that is, ``@obj.route`` is equivalent to
        ``@obj.route()``.

        :returns: Either the appropriately decorated method, or a
                  decorator for a method.
        """

        raise ValueError('cannot attach a method to a method')

    def mount(self, delegation, **kwargs):
        """
        Mount another controller class at this path element.  Only one
        delegation is allowed.

        :param delegation: Another controller to delegate to.  This
                           may be a ``micropath.Controller`` class
                           (not instance), or an instance of a
                           specialized subclass of ``Delegation``; the
                           latter possibility enables mounting other
                           WSGI applications, among other uses.

        :returns: The delegation to be mounted at this path element.
        :rtype: ``Delegation``

        :raises ValueError:
            The delegation has already been set.
        """

        # Method is overridden to ensure methods can't be added
        return super(Method, self).mount(delegation, **kwargs)


class Delegation(object):
    """
    Default delegation for controller classes.  Implements the
    descriptor protocol.  Subclasses should override the
    ``micropath_construct()`` method, which is passed the controller
    instance the delegation exists in and should return a new instance
    of the controller class passed to the constructor.
    """

    def __init__(self, controller, kwargs):
        """
        Initialize the ``Delegation`` instance.

        :param controller: A ``micropath.Controller`` class (not
                           instance) that will be delegated to.
        :param dict kwargs: Additional keyword arguments passed when
                            setting up the mount point.
        """

        # Save the controller and keyword arguments
        self.controller = controller
        self.kwargs = kwargs

        # The element may need to be added to the controller's root
        self.element = None

        # Cache of values
        self._cache = {}

    def __get__(self, obj, cls):
        """
        Implement the retriever portion of the descriptor protocol.

        :param obj: An instance of ``cls``.  If the value of the class
                    attribute is requested, this will be ``None``, and
                    this ``Delegation`` instance will be returned.
        :param cls: The class the class attribute is a member of.

        :returns: A properly constructed instance of the controller,
                  or this ``Delegation`` instance if ``obj`` is
                  ``None``.
        """

        # There are times we want to get to the Delegation
        if obj is None:
            return self

        return self.get(obj)

    def __set__(self, obj, value):
        """
        Set the value of the instance attribute.  This is provided for
        testing and debugging purposes.

        :param obj: An instance of the class the class attribute is a
                    member of.
        :param value: The value to assign to the instance attribute.
        """

        self._cache[id(obj)] = value

    def __delete__(self, obj):
        """
        Reset the value of the instance attribute.  This is provided for
        testing and debugging purposes.

        :param obj: An instance of the class the class attribute is a
                    member of.
        """

        # Try hard
        self._cache.pop(id(obj), None)

    def get(self, obj):
        """
        Retrieve a constructed instance of the class being delegated to.
        This calls the ``construct()`` method to construct a new
        instance of the class if necessary.

        :param obj: An instance of the class the class attribute is a
                    member of.

        :returns: An appropriately constructed instance of the target
                  class passed to the constructor.
        """

        key = id(obj)

        # Construct a new one if needed
        if key not in self._cache:
            self._cache[key] = self.construct(obj)

            # Set the parent and element
            self._cache[key]._micropath_parent = obj
            self._cache[key]._micropath_elem = self.element

        return self._cache[key]

    def construct(self, obj):
        """
        Construct a new instance of the controller class passed to the
        constructor.  This default implementation calls the
        ``micropath_construct()`` method, passing it the controller
        class and the keyword arguments set at mount time.

        :param obj: An instance of the class the class attribute is a
                    member of.

        :returns: An appropriately constructed instance of the target
                  class passed to the constructor.
        """

        return obj.micropath_construct(self.controller, self.kwargs)


def path(ident=None):
    """
    Construct a new ``Path`` element.

    :param str ident: The identifier for the component.  For constant
                      components, this will be the expected value of
                      the component.  May be ``None`` (the default) to
                      indicate that the metaclass should set the value
                      from the name of the class variable to which the
                      element is assigned.

    :returns: The newly constructed element.
    :rtype: ``Path``
    """

    return Path(ident)


def bind(ident=None):
    """
    Construct a new ``Binding`` element.

    :param str ident: The identifier for the component.  For variable
                      components, this will be the name of the
                      variable to contain the value.  May be ``None``
                      (the default) to indicate that the metaclass
                      should set the value from the name of the class
                      variable to which the element is assigned.

    :returns: The newly constructed element.
    :rtype: ``Binding``
    """

    # Construct the new binding
    return Binding(ident)


def route(*methods):
    """
    Method decorator to route HTTP methods to a method.  Called with
    an optional list of method strings (which will be canonicalized to
    uppercase).  If no methods are specified, all otherwise undefined
    methods will be routed to the decorated method.  This decorator is
    designed to work both with and without arguments--that is,
    ``@obj.route`` is equivalent to ``@obj.route()``.

    This decorator attaches a list of ``Method`` objects to the
    decorated function; these will be added to the controller class's
    root object appropriately by the metaclass.

    :returns: Either the appropriately decorated method, or a
              decorator for a method.
    """

    def decorator(func):
        meth_list = []

        # Construct the new Method objects
        seen = set()
        if methods:
            for meth_str in methods:
                if meth_str in seen:
                    continue
                seen.add(meth_str)

                meth = Method(meth_str, func)
                meth_list.append(meth)
        else:
            meth = Method(None, func)
            meth_list.append(meth)

        # Attach the method list to the function; this will be picked
        # up by the metaclass
        func._micropath_methods = meth_list

        # Mark the function as a handler
        func._micropath_handler = True

        # Pre-compute its want signature
        injector.WantSignature.from_func(func)

        return func

    # Check if methods consists of a single callable element
    if len(methods) == 1 and callable(methods[0]):
        func = methods[0]
        methods = ()
        return decorator(func)

    return decorator


def mount(delegation, *methods, **kwargs):
    """
    Mount another controller class at this path element.  Only one
    delegation is allowed.  The delegation can be restricted to a set
    of HTTP methods by listing the method name strings after the
    controller class; the default is to delegate all HTTP methods to
    the other controller.

    Additional keyword arguments will be passed when creating a
    ``Delegation``, which will subsequently pass them to the
    controller's ``micropath_construct()`` method.

    :param delegation: Another controller to delegate to.  This may be
                       a ``micropath.Controller`` class (not
                       instance), or an instance of a specialized
                       subclass of ``Delegation``; the latter
                       possibility enables mounting other WSGI
                       applications, among other uses.

    :returns: The delegation to be mounted at the root.
    :rtype: ``Delegation``
    """

    # Wrap the delegation, if necessary
    if not isinstance(delegation, Delegation):
        delegation = Delegation(delegation, kwargs)

    # If methods were specified, create them
    if methods:
        delegation._micropath_methods = []

        seen = set()
        for meth_str in methods:
            if meth_str in seen:
                continue
            seen.add(meth_str)

            meth = Method(meth_str, None)
            delegation._micropath_methods.append(meth)

            # Add the delegation to the Method
            meth.delegation = delegation

    return delegation



# Copyright (C) 2018 by Kevin L. Mitchell <klmitch@mit.edu>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License. You may
# obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied. See the License for the specific language governing
# permissions and limitations under the License.

from micropath import request


def safestr(value):
    """
    Turns ``None`` into the string "<None>".

    :param str value: The value to safely stringify.

    :returns: The stringified version of ``value``.
    """

    return value or '<None>'


def invoke(controller, *args, **kwargs):
    """
    Invoke a controller.  A request is constructed from the positional
    and keyword arguments.

    :param controller: An instance of the controller to invoke.
    :type controller: ``micropath.controller.Controller``

    :returns: A tuple of three elements.  The first element is the
              status string; the second element is a dictionary of
              headers; and the third is the textual contents of the
              body.
    """

    # First, construct a request
    req = request.Request.blank(*args, **kwargs)

    # Call the controller
    status, header_list, body_iter = req.call_application(controller)

    # Construct the header dictionary
    headers = {
        name.lower(): value for name, value in header_list
    }

    # Construct the body
    body = b''.join(body_iter)
    if hasattr(body_iter, 'close'):
        body_iter.close()

    return status, headers, body
